import numpy as np
import pytest
from astropy import units as u
from lightkurve import LightCurve

from lcviz import __version__, LCviz


@pytest.fixture
def helper():
    return LCviz()


@pytest.fixture
def light_curve_like_kepler_quarter(seed=42):
    """
    Generate a normalized (near unity) light curve with
    Gaussian noise. Times correspond very roughly to the
    Kepler 30-min cadences in Quarter 10. A random masking
    is applied to a small fraction of the fluxes.
    """
    np.random.seed(seed)
    exp_per_day = (30 * u.min).to_value(u.day)

    # approx start and stop JD for Kepler Quarter 10:
    time = np.arange(2455739, 2455833, exp_per_day)
    scale = 0.01
    flux = np.random.normal(
        1, scale=scale, size=len(time)
    ) * u.dimensionless_unscaled
    flux_err = scale * np.ones_like(flux)

    quality = np.zeros(len(time), dtype=np.int32)

    lc = LightCurve(
        time=time, flux=flux, flux_err=flux_err, quality=quality
    )
    lc['flux_alt'] = flux + 1
    lc['flux_alt_err'] = flux_err
    lc.meta.update(
        {
            'MISSION': 'KEPLER',
            'QUARTER': 10,
            'OBJECT': 'HAT-P-11',
            'RA': 297.7101763,
            'DEC': 48.0818635,
        }
    )

    return lc


try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
except ImportError:
    PYTEST_HEADER_MODULES = {}
    TESTED_VERSIONS = {}


def pytest_configure(config):
    PYTEST_HEADER_MODULES['astropy'] = 'astropy'
    PYTEST_HEADER_MODULES['glue-core'] = 'glue'
    PYTEST_HEADER_MODULES['glue-astronomy'] = 'glue_astronomy'
    PYTEST_HEADER_MODULES['gwcs'] = 'gwcs'
    PYTEST_HEADER_MODULES['lightkurve'] = 'lightkurve'

    TESTED_VERSIONS['lcviz'] = __version__



from traitlets import Bool, Unicode, observe
from lightkurve import LightCurve, LightCurveCollection

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetMultiSelectMixin,
                                        AddResultsMixin,
                                        with_spinner)
from jdaviz.core.user_api import PluginUserApi

from lcviz.utils import data_not_folded, is_not_tpf

__all__ = ['Stitch']


@tray_registry('stitch', label="Stitch")
class Stitch(PluginTemplateMixin, DatasetMultiSelectMixin, AddResultsMixin):
    """
    See the :ref:`Stitch Plugin Documentation <stitch>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Datasets to stitch.
    * ``remove_input_datasets``
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`stitch`
    """
    template_file = __file__, "stitch.vue"
    uses_active_status = Bool(False).tag(sync=False)

    remove_input_datasets = Bool(False).tag(sync=True)
    stitch_err = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._plugin_description = 'Stitch light curves together.'

        self.dataset.multiselect = True
        # do not support stitching data in phase-space
        # do not allow TPF as input
        self.dataset.add_filter(data_not_folded, is_not_tpf)

        self.results_label_default = 'stitched'

    @property
    def user_api(self):
        expose = ['dataset', 'stitch', 'remove_input_datasets', 'add_results']
        return PluginUserApi(self, expose=expose)

    @observe('dataset_items')
    def _set_relevent(self, *args):
        if len(self.dataset_items) < 2:
            self.irrelevant_msg = 'Requires at least two datasets loaded into viewers'
        else:
            self.irrelevant_msg = ''

    @with_spinner()
    def stitch(self, add_data=True):
        """
        Stitch multiple light curves (``dataset``) together using lightkurve.stitch.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting light curve to the app.

        Returns
        -------
        output_lc : `~lightkurve.LightCurve`
            The flattened light curve.
        """
        if not self.dataset.multiselect:
            raise ValueError("dataset must be in multiselect mode")
        if len(self.dataset.selected) < 2:
            raise ValueError("multiple datasets must be selected")
        lcc = LightCurveCollection([dci.get_object(LightCurve)
                                    for dci in self.dataset.selected_dc_item])
        stitched_lc = lcc.stitch(corrector_func=lambda x: x)

        if add_data:
            self.add_results.add_results_from_plugin(stitched_lc)
            if self.remove_input_datasets:
                for dataset in self.dataset.selected:
                    self.app.data_item_remove(dataset)
        return stitched_lc

    def vue_apply(self, *args, **kwargs):
        try:
            self.stitch(add_data=True)
        except Exception as e:
            self.stitch_err = str(e)
        else:
            self.stitch_err = ''



import pytest
import numpy as np
from numpy.testing import assert_allclose

from jdaviz.core.marks import MarkersMark


def _get_markers_from_viewer(viewer):
    return [m for m in viewer.figure.marks if isinstance(m, MarkersMark)][0]


def _assert_dict_allclose(dict1, dict2):
    assert dict1.keys() == dict2.keys()
    for k, v in dict1.items():
        if isinstance(v, float):
            assert_allclose(v, dict2.get(k))
        elif isinstance(v, (tuple, list)):
            assert_allclose(np.asarray(v), np.asarray(dict2.get(k)))
        else:
            assert v == dict2.get(k)


def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load_data(lc)
    # lcviz.show()

    markers = lcviz.plugins['Markers']
    markers.open_in_tray()
    # interactively mark by mousing over the viewer and pressing "M"
    table = markers.export_table()
    print(table)
    markers.clear_table()


def test_plugin_markers(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    tv = helper.default_time_viewer._obj

    mp = helper.plugins['Markers']
    label_mouseover = mp._obj.coords_info
    mp.open_in_tray()

    # test event in flux-vs-time viewer
    label_mouseover._viewer_mouse_event(tv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})

    assert label_mouseover.as_text() == ('Cursor 0.00000e+00, 0.00000e+00',
                                         'Time 5.45833e+00 d',
                                         'Flux 9.67587e-01')

    object_name = helper.app.data_collection[-1].meta['OBJECT']
    _assert_dict_allclose(label_mouseover.as_dict(), {'data_label': f'{object_name} [Q10]',
                                                      'time': 5.4583335,
                                                      'time:unit': 'd',
                                                      'phase': np.nan,
                                                      'ephemeris': '',
                                                      'axes_x': 5.4583335,
                                                      'axes_x:unit': 'd',
                                                      'index': 262.0,
                                                      'axes_y': 0.96758735,
                                                      'axes_y:unit': '',
                                                      'value': 0.96758735,
                                                      'value:unit': ''})

    mp._obj._on_viewer_key_event(tv, {'event': 'keydown',
                                      'key': 'm'})
    assert len(mp.export_table()) == 1
    assert len(_get_markers_from_viewer(tv).x) == 1

    ephem = helper.plugins['Ephemeris']
    pv = ephem.create_phase_viewer()._obj

    # test event in flux-vs-phase viewer
    label_mouseover._viewer_mouse_event(pv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0.5, 'y': 0}})

    assert label_mouseover.as_text() == ('Cursor 5.00000e-01, 0.00000e+00',
                                         'Phase 0.45833',
                                         'Flux 9.67587e-01')

    _assert_dict_allclose(label_mouseover.as_dict(), {'data_label': f'{object_name} [Q10]',
                                                      'time': 5.458333374001086,
                                                      'time:unit': 'd',
                                                      'phase': 0.4583333730697632,
                                                      'ephemeris': 'default',
                                                      'axes_x': 0.4583333730697632,
                                                      'axes_x:unit': '',
                                                      'index': 262.0,
                                                      'axes_y': 0.9675873517990112,
                                                      'axes_y:unit': '',
                                                      'value': 0.9675873517990112,
                                                      'value:unit': ''})

    mp._obj._on_viewer_key_event(pv, {'event': 'keydown',
                                      'key': 'm'})
    assert len(mp.export_table()) == 2
    assert len(_get_markers_from_viewer(tv).x) == 1
    assert len(_get_markers_from_viewer(pv).x) == 1

    # test event in flux-vs-phase viewer (with cursor only)
    label_mouseover.dataset.selected = 'none'
    label_mouseover._viewer_mouse_event(pv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0.6, 'y': 0}})

    print(label_mouseover.as_text())
    assert label_mouseover.as_text() == ('Cursor 6.00000e-01, 0.00000e+00',
                                         '',
                                         '')

    _assert_dict_allclose(label_mouseover.as_dict(), {'axes_x': 0.6,
                                                      'axes_x:unit': '',
                                                      'axes_y': 0,
                                                      'axes_y:unit': '',
                                                      'data_label': '',
                                                      'time': np.nan,
                                                      'time:unit': '',
                                                      'phase': 0.6,
                                                      'value': 0,
                                                      'value:unit': '',
                                                      'ephemeris': ''})


@pytest.mark.remote_data
def test_tpf_markers(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)

    # TODO: replace with test fixture
    from lightkurve import search_targetpixelfile
    tpf = search_targetpixelfile("KIC 001429092",
                                 mission="Kepler",
                                 cadence="long",
                                 quarter=10).download()
    helper.load_data(tpf)

    mp = helper.plugins['Markers']
    label_mouseover = mp._obj.coords_info
    mp.open_in_tray()

    assert abs(helper.plugins['Time Selector'].value - 47.006888) < 1e-4

    # test event in image (TPF) viewer
    iv = helper.viewers['image']._obj
    label_mouseover._viewer_mouse_event(iv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})

    assert label_mouseover.as_text() == ('Pixel x=00000.0 y=00000.0 Value +1.28035e+01 electron / s',  # noqa
                                         'Time 47.00689 d',
                                         '')

    print(label_mouseover.as_dict())
    _assert_dict_allclose(label_mouseover.as_dict(), {'data_label': 'KIC 1429092[TPF]',
                                                      'time': 47.00689,
                                                      'time:unit': 'd',
                                                      'pixel': (0.0, 0.0),
                                                      'axes_x': 0,
                                                      'axes_x:unit': 'pix',
                                                      'axes_y': 0,
                                                      'axes_y:unit': 'pix',
                                                      'value': 12.803529,
                                                      'value:unit': 'electron / s'})



import pytest


def test_reset_limits(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    tv = helper.default_time_viewer._obj

    orig_xlims = (tv.state.x_min, tv.state.x_max)
    orig_ylims = (tv.state.y_min, tv.state.y_max)
    # set xmin and ymin to midpoints
    new_xmin = (tv.state.x_min + tv.state.x_max) / 2
    new_ymin = (tv.state.y_min + tv.state.y_max) / 2
    tv.state.x_min = new_xmin
    tv.state.y_min = new_ymin

    tv.state._reset_x_limits()
    assert tv.state.x_min == orig_xlims[0]
    assert tv.state.y_min == new_ymin

    tv.state._reset_y_limits()
    assert tv.state.y_min == orig_ylims[0]


@pytest.mark.remote_data
def test_clone(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)

    def_viewer = helper.viewers['flux-vs-time']
    assert helper._get_clone_viewer_reference(def_viewer._obj.reference) == 'flux-vs-time[1]'

    new_viewer = def_viewer._obj.clone_viewer()
    assert helper._get_clone_viewer_reference(new_viewer._obj.reference) == 'flux-vs-time[2]'

    # TODO: replace with test fixture
    from lightkurve import search_targetpixelfile
    tpf = search_targetpixelfile("KIC 001429092",
                                 mission="Kepler",
                                 cadence="long",
                                 quarter=10).download()
    helper.load_data(tpf)
    im_viewer = helper.viewers['image']
    assert helper._get_clone_viewer_reference(im_viewer._obj.reference) == 'image[1]'
    im_viewer._obj.clone_viewer()



import os
from glue.core import DataCollection
from glue.core.roi import XRangeROI
import numpy as np
from astropy.time import Time
from lightkurve import search_lightcurve, LightCurve


SMALL_LC_PATH = os.path.join(
    os.path.dirname(__file__), os.pardir, 'data',
    'HAT-P-11_Kepler_Q0_snippet.fits'
)


def test_translator(light_curve_like_kepler_quarter):
    light_curve = light_curve_like_kepler_quarter
    dc = DataCollection()
    dc['dummy-lc'] = light_curve
    translated_lc = dc['dummy-lc'].get_object()

    # check for equality within these attrs:
    attrs = ['time', 'flux', 'flux_err']

    for attr in attrs:
        expected = getattr(light_curve, attr)
        translated = getattr(translated_lc, attr)

        if attr == 'time':
            args = (
                expected.tdb.jd,
                translated.tdb.jd
            )
        else:
            args = (
                expected.value,
                translated.to_value(expected.unit)
            )

        # check that attribute's values match expectations:
        np.testing.assert_allclose(*args)

        if hasattr(expected, 'mask'):
            # check that mask is preserved
            assert np.all(expected.mask == translated.mask)


def _download_and_write_lc_snippet():
    """
    Download and write a local copy of the first 10 long cadences of the
    Kepler light curve of HAT-P-11 in Quarter 0.

    Run it with:
    python -c "from lcviz.tests.test_translator import _download_and_write_lc_snippet as d; d()"
    """
    lc = search_lightcurve(
        "HAT-P-11", mission="Kepler", cadence="long", quarter=0
    ).download()
    lc[0:10].write(SMALL_LC_PATH, overwrite=True)


def test_round_trip(helper):
    """
    The FITS file used in this test is generated by the
    function ``_download_and_write_lc_snippet``.
    """
    lc = LightCurve.read(SMALL_LC_PATH, astropy_native=True)
    helper.load_data(lc)

    # these times chosen by hand during Kepler Q0:
    near_transit = Time([
        '2009-05-02 02:51',
        '2009-05-02 03:52'
    ], format='iso')

    viewer = helper.default_time_viewer._obj
    viewer.apply_roi(XRangeROI(*near_transit))

    columns_to_check = ['time', 'flux', 'flux_err']
    subset_translated = helper.get_data(subset='Subset 1')[columns_to_check]
    row_mask = (lc.time > near_transit[0]) & (lc.time < near_transit[1])
    lc_subset = lc[row_mask][columns_to_check]

    for column in lc_subset.values_equal(subset_translated).itercols():
        assert np.all(column)



from glue.core.subset import Subset
from glue.config import data_translator
from glue.core import BaseData
from glue.core.exceptions import IncompatibleAttribute
from glue.core.roi import RangeROI
from glue.core.subset_group import GroupedSubset

from glue_jupyter.bqplot.scatter import BqplotScatterView

from astropy import units as u
from astropy.time import Time

from jdaviz.core.events import NewViewerMessage
from jdaviz.core.registries import viewer_registry
from jdaviz.configs.cubeviz.plugins.viewers import (CubevizImageView,
                                                    WithSliceIndicator, WithSliceSelection)
from jdaviz.configs.default.plugins.viewers import JdavizViewerMixin
from jdaviz.configs.specviz.plugins.viewers import SpecvizProfileView

from lcviz.state import ScatterViewerState

from lightkurve import LightCurve

__all__ = ['TimeScatterView', 'PhaseScatterView', 'CubeView']


class CloneViewerMixin:
    def _get_clone_viewer_reference(self):
        base_name = self.reference.split("[")[0]
        name = base_name
        ind = 0
        while name in self.jdaviz_helper.viewers.keys():
            ind += 1
            name = f"{base_name}[{ind}]"
        return name

    def clone_viewer(self):
        name = self.jdaviz_helper._get_clone_viewer_reference(self.reference)

        self.jdaviz_app._on_new_viewer(NewViewerMessage(self.__class__,
                                                        data=None,
                                                        sender=self.jdaviz_app),
                                       vid=name, name=name)

        nv = self.jdaviz_helper.viewers.get(name)

        visible_layers = self.data_menu.data_labels_visible
        for layer in self.data_menu.data_labels_loaded:
            visible = layer in visible_layers
            nv.data_menu.add_data(layer)
            nv.data_menu.set_layer_visibility(layer, visible)
            # TODO: don't revert color when adding same data to a new viewer
            # (same happens when creating a phase-viewer from ephemeris plugin)

        new_viewer = self.jdaviz_app.get_viewer(name)
        if hasattr(self, 'ephemeris_component'):
            new_viewer._ephemeris_component = self._ephemeris_component
        for k, v in self.state.as_dict().items():
            if k in ('layers',):
                continue
            setattr(new_viewer.state, k, v)

        for this_layer_state, new_layer_state in zip(self.state.layers, new_viewer.state.layers):
            for k, v in this_layer_state.as_dict().items():
                if k in ('layer',):
                    continue
                setattr(new_layer_state, k, v)

        return new_viewer.user_api


@viewer_registry("lcviz-time-viewer", label="flux-vs-time")
class TimeScatterView(JdavizViewerMixin, CloneViewerMixin, WithSliceIndicator, BqplotScatterView):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom', 'jdaviz:xrangezoom', 'jdaviz:yrangezoom'],
                    ['jdaviz:panzoom', 'jdaviz:panzoom_x', 'jdaviz:panzoom_y'],
                    ['bqplot:xrange', 'bqplot:yrange', 'bqplot:rectangle'],
                    ['jdaviz:selectslice'],
                    ['lcviz:viewer_clone', 'jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]
    default_class = LightCurve
    _state_cls = ScatterViewerState

    _native_mark_classnames = ('Image', 'ImageGL', 'Scatter', 'ScatterGL')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.display_mask = False
        self.time_unit = kwargs.get('time_unit', u.d)
        self.initialize_toolbar(default_tool_priority=['jdaviz:selectslice'])
        self._subscribe_to_layers_update()
        # hack to inherit a small subset of methods from SpecvizProfileView
        # TODO: refactor jdaviz so these can be included in some mixin
        self._show_uncertainty_changed = lambda value: SpecvizProfileView._show_uncertainty_changed(self, value)  # noqa
        self._plot_uncertainties = lambda: SpecvizProfileView._plot_uncertainties(self)
        # TODO: _plot_uncertainties in specviz is hardcoded to look at spectral_axis and so crashes
        self._clean_error = lambda: SpecvizProfileView._clean_error(self)
        self.density_map = kwargs.get('density_map', False)

    @property
    def slice_component_label(self):
        # label of the component in the lightcurves corresponding to the slice axis
        # calling data_collection_item.get_component(slice_component_label) must work
        return 'dt'

    @property
    def slice_display_unit_name(self):
        return 'time'

    def data(self, cls=None):
        data = []

        # TODO: generalize upstream in jdaviz.
        # This method is generalized from
        # jdaviz/configs/specviz/plugins/viewers.py
        # to support non-spectral viewers.
        for layer_state in self.state.layers:
            if hasattr(layer_state, 'layer'):
                lyr = layer_state.layer
                # For raw data, just include the data itself
                if isinstance(lyr, BaseData):
                    _class = cls or self.default_class

                    if _class is not None:
                        cache_key = lyr.label
                        if cache_key in self.jdaviz_app._get_object_cache:
                            layer_data = self.jdaviz_app._get_object_cache[cache_key]
                        else:
                            layer_data = lyr.get_object(cls=_class)
                            self.jdaviz_app._get_object_cache[cache_key] = layer_data

                        data.append(layer_data)

                # For subsets, make sure to apply the subset mask to the layer data first
                elif isinstance(lyr, (Subset, GroupedSubset)):
                    layer_data = lyr

                    if _class is not None:
                        handler, _ = data_translator.get_handler_for(_class)
                        try:
                            layer_data = handler.to_object(layer_data)
                        except IncompatibleAttribute:
                            continue
                    data.append(layer_data)

        return data

    def _apply_layer_defaults(self, layer_state):
        if getattr(layer_state.layer, 'meta', {}).get('Plugin', None) == 'Binning':
            # increased size of binned results, by default
            layer_state.size = 5
        layer_state.points_mode = 'markers'

    def set_plot_axes(self):
        # set which components should be plotted
        dc = self.jdaviz_app.data_collection
        component_labels = [comp.label for comp in dc[0].components]

        # Get data to be used for axes labels
        light_curve = self.data()[0]
        self._set_plot_x_axes(dc, component_labels, light_curve)
        self._set_plot_y_axes(dc, component_labels, light_curve)

    def _set_plot_x_axes(self, dc, component_labels, light_curve=None, reference_time=None):
        self.state.x_att = dc[0].components[component_labels.index('dt')]

        x_unit = self.time_unit

        if light_curve is not None and reference_time is None:
            reference_time = light_curve.meta.get('reference_time', None)
        elif reference_time is None:
            reference_time = dc[0].coords.reference_time

        xlabel = f'{str(x_unit.physical_type).title()} from {reference_time.iso} ({x_unit})'

        self.figure.axes[0].label = xlabel
        self.figure.axes[0].num_ticks = 5

    def _set_plot_y_axes(self, dc, component_labels, light_curve):
        self.state.y_att = dc[0].components[component_labels.index('flux')]

        y_unit = light_curve.flux.unit
        y_unit_physical_type = str(y_unit.physical_type).title()

        common_count_rate_units = (u.electron / u.s, u.dn / u.s, u.ct / u.s)

        if y_unit_physical_type == 'Unknown':
            if y_unit.is_equivalent(common_count_rate_units):
                y_unit_physical_type = 'Flux'
        if y_unit_physical_type == 'Dimensionless':
            y_unit_physical_type = 'Relative Flux'

        ylabel = f'{y_unit_physical_type}'

        if not y_unit.is_equivalent(u.dimensionless_unscaled):
            ylabel += f' ({y_unit})'

        self.figure.axes[1].label = ylabel

        # Make it so y axis label is not covering tick numbers (sometimes)
        self.figure.axes[1].label_offset = "-50"

        # Set (X,Y)-axis to scientific notation if necessary:
        self.figure.axes[0].tick_format = 'g'
        self.figure.axes[1].tick_format = 'g'

        self.figure.axes[1].num_ticks = 5

    def _expected_subset_layer_default(self, layer_state):
        super()._expected_subset_layer_default(layer_state)

        layer_state.linewidth = 3

        # optionally prevent subset from being rendered
        # as a density map, rather than shaded markers over data:
        layer_state.density_map = self.density_map

    def add_data(self, data, color=None, alpha=None, **layer_state):
        """
        Overrides the base class to handle subset styling defaults.

        Parameters
        ----------
        data : :class:`glue.core.data.Data`
            Data object with the light curve.
        color : obj
            Color value for plotting.
        alpha : float
            Alpha value for plotting.

        Returns
        -------
        result : bool
            `True` if successful, `False` otherwise.
        """
        result = super().add_data(data, color, alpha, **layer_state)

        for layer in self.layers:
            # optionally render as a density map
            layer.state.density_map = self.density_map

        # Set default linewidth on any created subset layers
        for layer in self.state.layers:
            if "Subset" in layer.layer.label and layer.layer.data.label == data.label:
                layer.linewidth = 3

        # update viewer limits when data are added
        self.set_plot_axes()
        self.state.reset_limits()

        return result

    def _show_uncertainty_changed(*args, **kwargs):
        # method required by jdaviz
        pass

    def apply_roi(self, roi, use_current=False):
        if isinstance(roi, RangeROI):
            # allow ROIs describing times to be applied with min and max defined as:
            #  1. floats, representing bounds in units of ``self.time_unit``
            #  2. Time objects, which get converted to work like (1) via the reference time
            if isinstance(roi.min, Time) or isinstance(roi.max, Time):
                reference_time = self.data()[0].meta.get('reference_time', 0)
                roi = roi.transformed(xfunc=lambda x: (x - reference_time).to_value(self.time_unit))

        super().apply_roi(roi, use_current=use_current)


@viewer_registry("lcviz-phase-viewer", label="flux-vs-phase")
class PhaseScatterView(TimeScatterView):
    def __init__(self, *args, **kwargs):
        self._ephemeris_component = 'default'
        super().__init__(*args, **kwargs)

    @property
    def ephemeris(self):
        ephem = self.jdaviz_helper.plugins.get('Ephemeris', None)
        if ephem is None:
            raise ValueError("must have ephemeris plugin loaded to access ephemeris")
        return ephem.ephemerides.get(self._ephemeris_component)

    def _set_plot_x_axes(self, dc, component_labels, light_curve):
        # setting of y_att will be handled by ephemeris plugin
        self.state.x_att = dc[0].components[component_labels.index(f'phase:{self._ephemeris_component}')]  # noqa
        self.figure.axes[0].label = 'phase'
        self.figure.axes[0].num_ticks = 5

    def times_to_phases(self, times):
        ephem = self.jdaviz_helper.plugins.get('Ephemeris', None)
        if ephem is None:
            raise ValueError("must have ephemeris plugin loaded to convert")

        return ephem.times_to_phases(times, ephem_component=self._ephemeris_component)

    def _set_slice_indicator_value(self, value):
        # NOTE: on first call, this will initialize the indicator itself
        self.slice_indicator.value = self.times_to_phases(value)


@viewer_registry("lcviz-cube-viewer", label="cube")
class CubeView(CloneViewerMixin, CubevizImageView, WithSliceSelection):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom'],
                    ['jdaviz:panzoom'],
                    ['bqplot:rectangle'],
                    ['lcviz:viewer_clone', 'jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]
    # TODO: can we vary this default_class based on Kepler vs TESS, etc?
    # see https://github.com/spacetelescope/lcviz/pull/81#discussion_r1469721009
    default_class = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.display_mask = False
        self.time_unit = kwargs.get('time_unit', u.d)
        self.initialize_toolbar()
        self._subscribe_to_layers_update()

        # Hide axes by default
        self.state.show_axes = False

        # TODO: refactor upstream so lcviz can inherit cubeviewer methods/setup without
        # jdaviz-specific logic:
        # * _default_spectrum_viewer_reference_name
        # * _default_flux_viewer_reference_name
        # * _default_uncert_viewer_reference_name

    @property
    def slice_component_label(self):
        # label of the component in the cubes corresponding to the slice axis
        # calling data_collection_item.get_component(slice_component_label) on any
        # input cube-data must work
        return 'dt'

    @property
    def slice_index(self):
        # index in viewer.slices corresponding to the slice axis
        return 0

    @property
    def slice_display_unit_name(self):
        return 'time'

    def _initial_x_axis(self, *args):
        # Make sure that the x_att/y_att is correct on data load
        # called via a callback set upstream in CubevizImageView when reference_data is changed
        ref_data = self.state.reference_data
        if ref_data is not None:
            self.state.x_att = ref_data.id['Pixel Axis 2 [x]']
            self.state.y_att = ref_data.id['Pixel Axis 1 [y]']

    def _on_layers_update(self, layers=None):
        super()._on_layers_update(layers=layers)
        ref_data = self.state.reference_data
        if ref_data is None:
            return
        flux_comp = ref_data.id['flux']
        for layer in self.state.layers:
            if hasattr(layer, 'attribute') and layer.attribute != flux_comp:
                layer.attribute = flux_comp

    def data(self, cls=None):
        # TODO: generalize upstream in jdaviz.
        # This method is generalized from
        # jdaviz/configs/cubeviz/plugins/viewers.py
        return [layer_state.layer
                for layer_state in self.state.layers
                if hasattr(layer_state, 'layer') and
                isinstance(layer_state.layer, BaseData)]
